open Term
open Trs
open Monad.MonadList

(* TODO: Weakly Orthogonal Composability *)

(** constructor sharing **)

let is_cons_share r s =
  Listset.subset
    (Listset.inter (Trs.functions r) (Trs.functions s))
    (Listset.inter (Trs.conss r) (Trs.conss s))
  
let all_cons_share rss =
  List.for_all
    (fun (r,s) -> is_cons_share r s)
    (Listx.tri rss)


(** composability **)
(* restriction *)
let restrict rs p =
  List.filter
    (fun (l,r) -> List.mem (root l) p)
    rs

let rec composable r s = cond1 r s && cond1 s r && cond2 r s
and cond1 r s = Listset.inter (conss r) (defs s) = []
and cond2 r s =
  Listset.subset
    (restrict (r@s) (Listset.inter (defs r) (defs s)))
    (Listset.inter r s)

let all_composable rss =
  List.for_all
    (fun (r,s) -> composable r s)
    (Listx.tri rss)

let cond2WMO r s = cond2 r s && Match.is_WMO r s
let composableWMO r s = composable r s && cond2WMO r s
let all_composableWMO rss =
  List.for_all
    (fun (r,s) -> composableWMO r s)
    (Listx.tri rss)

let commuting r s = Result.is_YES (Commute.direct r s)
let all_commuting rss =
  List.for_all
    (fun (r,s) -> commuting r s)
    (Listx.tri rss)

(********* DECOMPOSITION **********)

(* if Cr and Ds have common symbols A, move rules in S to R
     s.t. defined symbols of Rule.t are included in A
 * suppose Cr and Ds have common symbol *)
let fix_cond1 r s =
  let is = restrict s (Listset.inter (conss r) (defs s)) in
  (Listset.union is r, s)

(* if R has rules that has common defined symbol on R and S, and
     these rules are not contained in S, then move these rules to S *)
let fix_cond2 r s =
  let is = restrict (Listset.union r s) (Listset.inter (defs r) (defs s)) in
  (Listset.union is r, Listset.union is s)

(* suppose R and S are composable *)
let fix_WMO r s =
  if Match.is_WMO r s then
    [r;s]
  else
    [Listset.union r s]
let fix_COM r s =
  if commuting r s then
    [r;s]
  else
    [Listset.union r s]

let subset x y = 
  not (Listset.equal x y) && Listset.subset x y

let rec remove_subset = function
  | []    -> []
  | [x]   -> [x]
  | x::xs ->
      let ys =
        List.filter (fun y -> not (Listset.subset y x)) xs in
      if List.exists (fun y -> subset x y) xs then
        remove_subset ys
      else
        x :: remove_subset ys

let small_order xss =
  let len = List.length in
  List.sort
    (fun xs ys -> compare (len xs) (len ys))
    xss

(* merge subsystems in rss s.t. overlapping function and defined symbols *)
let rec members rss rs =
  let non_css =
    List.filter
      (fun ss -> not (is_cons_share rs ss))
      rss in
  match non_css with
  | []  -> rs
  | sss -> members
             (Listset.diff rss sss)
             (Listset.union rs (Listset.big_union sss))

(* merge all subsystems by function 'member' *)
let rec member_class = function
  | []      -> []
  | rs::rss ->
      let m = members rss rs in
      let sss =
        List.filter
          (fun ss -> not (Listset.subset ss m))
          rss in
      m :: member_class sss

let decompose_cons_share rs =
  let ds = List.map (fun (l,r) -> root l, (l,r)) rs in
  let rec loop = function
      []       -> []
    | r :: xss ->
       match List.filter (fun s -> not (is_cons_share r s)) xss with
         []  -> r :: loop xss
       | yss -> let rs = Listset.big_union (r :: yss) in
                loop (rs :: Listset.diff xss (r :: yss))
  in loop (List.map snd (Listx.group ds))
  
let cons_share rs =
  let singletons =
    List.map Listset.singleton rs in
  small_order @@ member_class singletons

let cons_share rs = Console.watcher_l
                      "Constructor-sharing decomposition"
                      cons_share
                      rs

let rec decompose rs =
  small_order @@ decompose' @@
    (* initial TRSs are grouped by defined symbol *)
    (* group_by (fun a b -> def a = def b) rs *)
    decompose_cons_share rs
and decompose' rss =
  if all_composable rss then
    rss
  else
    let sss =
      let* (r,s) = Listx.tri rss in
      let (r',s') = proc r s in
      [r'; s'] in
    decompose' @@ remove_subset @@ sss
and proc r s =
  match cond1 r s, cond1 s r, cond2 r s with
  | false,_,_ -> fix_cond1 r s
  | _,false,_ -> fix_cond1 s r
  | _,_,false -> fix_cond2 r s
  | _         -> (r,s)

let rec decomposeWMO rs =
  small_order @@ decomposeWMO' (decompose rs)
and decomposeWMO' rss =
  if all_composableWMO rss then
    rss
  else
    let sss =
      Listx.concat_map
        (fun (r,s) -> fix_WMO r s)
        (Listx.tri rss) in
    decomposeWMO' @@ remove_subset @@ sss

let rec decomposeCOM rs =
  small_order @@ decomposeCOM' (decompose rs)
and decomposeCOM' rss =
  if all_commuting rss then
    rss
  else
    let sss =
      Listx.concat_map
        (fun (r,s) -> fix_COM r s)
        (Listx.tri rss) in
    decomposeCOM' @@ remove_subset @@ sss

let decompose rs =
  Console.watcher_l "Composable decomposition" decompose rs
let decomposeWMO rs =
  Console.watcher_l "Composable decompositionWMO" decomposeWMO rs
let decomposeCOM rs =
  Console.watcher_l "Composable decompositionCOM" decomposeCOM rs


(** Utility **)
let is_composable rs = List.length (decompose rs) > 1
let is_cons_share rs = List.length (cons_share rs) > 1

let show_decompose rss =
  Format.printf "%d decompositions@." (List.length rss);
  List.iter
    (fun (i,rs) ->
      Format.printf " #%d -----------\n%s@."
        i
        (sprint_rules rs))
    (Listx.ix rss)

let print_decompose  rs = show_decompose (decompose rs)
let print_cons_share rs = show_decompose (cons_share rs)
