(*
including Applicative (return) and Alternative (empty, <|>)
module type Monad = sig
  type 'a t
  val return : 'a -> 'a t
  val empty : 'a t
  val (<|>) : 'a t -> 'a t -> 'a t
  val bind : 'a t -> ('a -> 'b t) -> 'b t
  val ( let* ) : 'a t -> ('a -> 'b t) -> 'b t
  val (>>=) : 'a t -> ('a -> 'b t) -> 'b t
  val (>>) : ('a -> 'b t) -> ('b -> 'c t) -> 'a -> 'c t
  val guard : bool -> () t
end
*)

(* usage:

  let* x = [1;2;3] in return (x*1)

equals to 

  [x*1 | x <- [1;2;3]]

also

  let* x = [1;2;3;4;5;6] in
  guard (1<x) >>= fun () ->
  let* y = [x*2; x*4] in
  guard (y<9) >>= fun () ->
  return y

is

  [x*2 | x <- [1;2;3;4;5;6], 1<x,
         y <- [x*2; x*4],    x<6]
*)

module MonadList = struct
  let return x = [x]
  let empty = []
  let (<|>) = (@)
  let bind xs f = List.concat (List.map f xs)
  let (let*) = bind
  let (>>=) = bind
  let (>>) f g = fun xs -> f xs >>= g
  let guard b = if b then return () else empty
end


module MonadOption = struct
  let return x = Some x
  let empty = None
  let (<|>) l r = match (l,r) with (Some _,_) -> l | _ -> r
  let bind = Option.bind
  let ( let* ) = bind
  let (>>=) = bind
  let (>>) f g = fun x -> match f x with None -> None | Some y -> g y
  let guard b = if b then return () else empty
end
